use x86::io::{outb, inb};

use spin::Mutex;

#[macro_use]
pub mod macros;

lazy_static::lazy_static!{
    pub static ref WRITER: Mutex<Writer> = Mutex::new(Writer::new());
}

const BUFFER_HEIGHT: usize = 25;
const BUFFER_WIDTH: usize = 80;

#[allow(dead_code)]
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[repr(u8)]
pub enum Color {
    Black = 0,
    Blue = 1,
    Green = 2,
    Cyan = 3,
    Red = 4,
    Magenta = 5,
    Brown = 6,
    LightGray = 7,
    DarkGray = 8,
    LightBlue = 9,
    LightGreen = 10,
    LightCyan = 11,
    LightRed = 12,
    Pink = 13,
    Yellow = 14,
    White = 15,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[repr(transparent)]
pub struct ColorCode(u8);

impl ColorCode {
    pub const fn new(foreground: Color, background: Color) -> ColorCode {
        ColorCode((background as u8) << 4 | (foreground as u8))
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[repr(C)]
pub struct VgaChar {
    pub chr: u8,
    pub color: ColorCode,
}


#[repr(transparent)]
pub struct Buffer {
    chars: [[VgaChar; BUFFER_WIDTH]; BUFFER_HEIGHT],
}

pub struct Writer {
    col: usize,
    line: usize,
    buffer: &'static mut Buffer,
    color: ColorCode
}

impl Writer {
    pub fn new() -> Self {
        let mut s = Self {
            col: 0,
            line: 0,
            buffer: unsafe { &mut *(0xb8000 as *mut Buffer) },
            color: ColorCode::new(Color::White, Color::Black)
        };
        s.clear_screen();
        s
    }

    fn write_byte(&mut self, chr: u8) {
        match chr {
            b'\n' => self.new_line(),
            _ => {
                if self.col >= BUFFER_WIDTH {
                    self.new_line();
                }

                self.buffer.chars[self.line][self.col] = VgaChar {
                    chr,
                    color: self.color,
                };
                self.col += 1;   
            }
        }
    }

    pub fn write_string(&mut self, s: &str) {
        for c in s.as_bytes() {
            match c {
                // printable ASCII byte or newline
                0x20..=0x7e | b'\n' => self.write_byte(*c),
                // not part of printable ASCII range
                _ => self.write_byte(0xfe),
            }
        }
        update_cursor(self.col as u16, self.line as u16);
    }

    
    #[allow(dead_code)]
    pub fn set_color(&mut self, fg: Color, bg: Color) {
        self.color = ColorCode::new(fg, bg);
    }

    pub fn set_color_code(&mut self, c: ColorCode) {
        self.color = c;
    }
    
    pub fn clear_screen(&mut self) {
        let color = ColorCode::new(Color::White, Color::Black);
        for line in 0..BUFFER_HEIGHT {
            for col in 0..BUFFER_WIDTH {
                self.buffer.chars[line][col] = VgaChar {
                    chr: 0,
                    color,
                };
            }
        }
    }

    pub fn clear_line(&mut self, line: usize) {
        let color = ColorCode::new(Color::White, Color::Black);
        for col in 0..BUFFER_WIDTH {
            self.buffer.chars[line][col] = VgaChar {
                chr: 0,
                color,
            };
        }
    }

    pub fn new_line(&mut self) {
        if self.line >= BUFFER_HEIGHT - 1{ 
            for row in 1..BUFFER_HEIGHT {
                for col in 0..BUFFER_WIDTH {
                    let chr = self.buffer.chars[row][col];
                    self.buffer.chars[row - 1][col] = chr;
                }
            }
            self.clear_line(BUFFER_HEIGHT - 1);
        } else {
            self.line += 1;
        }
        self.col = 0;

    }
}

#[allow(dead_code)]
pub fn enable_cursor(cur_start: u8, cur_end: u8) {
    unsafe {
        outb(0x3D4, 0x0A);
        outb(0x3D5, (inb(0x3D5) & 0xC0) | cur_start);
        
        outb(0x3D4, 0x0B);
        outb(0x3D5, (inb(0x3D5) & 0xE0) | cur_end);
    }
}

#[allow(dead_code)]
pub fn disable_cursor() {
    unsafe {
        outb(0x3D4, 0x0A);
        outb(0x3D5, 0x20);
    }
}

pub fn update_cursor(x: u16, y: u16) {
    unsafe {

        let pos: u16 = y * ( BUFFER_WIDTH as u16) + x;
        
        outb(0x3D4, 0x0F);
        outb(0x3D5, (pos & 0xFF) as u8);
        outb(0x3D4, 0x0E);
        outb(0x3D5, ((pos >> 8 ) & 0xFF) as u8);
    }
}

#[allow(dead_code)]
pub fn get_cursor_position() -> (u16, u16) {
    let mut pos: u16 = 0;
    unsafe {

        outb(0x3D4, 0x0F);
        pos |= inb(0x3D5) as u16;
        outb(0x3D4, 0x0E);
        pos |= (inb(0x3D5) as u16) << 8;
    }
    (pos / (BUFFER_WIDTH as u16), pos % (BUFFER_WIDTH as u16))
}
